︠8b38344f-19fe-4546-8969-ccb34ed2e93ai︠
%html
<h1><center>Elliptic Curve Procedures</center></h1>
︡90129acb-5ecb-4194-85aa-b2f70aa1a88a︡{"done":true,"html":"<h1><center>Elliptic Curve Procedures</center></h1>"}
︠366c869c-a86c-443c-82cc-8c634b35e6efs︠


import numpy
import itertools

############################################################################################################################
# For all these procedures we are working with elliptic curve (EC) groups over Z_p where p is a prime number larger than 3.#
# An Elliptic Curve group will be represented by a triple [A,B,p] where                                                    #
#      (1) p is a prime number larger than 3                                                                               #
#      (2) A is the coefficient of x and                                                                                   #
#      (3) B is the constant term in  y^2 = x^3 + A*x + B mod p.                                                           #
#                                                                                                                          #
# A point in the group will be represented as a pair [x,y] where                                                           #
#      (1)  x is the x-coordinate of a solution                                                                            #
#      (2)  y is the y-coordinate of a solution                                                                            #
#      (3)  The identity will be represented by []                                                                         #
############################################################################################################################

def TonSh (a, Prime):
    if kronecker(a, Prime) == -1:
        pass
        print("{0} does not have a square root mod {1}".format(a, Prime))
        return None
    elif Prime % 4 == 3:
        x = power_mod(ZZ(a), ZZ((Prime+1)/4),ZZ(Prime))
        return(x)
    else:
        #################################################################################
        # Determine e so that Prime-1 = (2^e)*q for some odd number q                   #
        #################################################################################

        e = ZZ(0)
        q = ZZ(Prime - 1)
#        for j in range(1, Prime):
        for j in itertools.count(1):
            if q % 2 == 0:
                e = ZZ(e + 1)
                q = ZZ(q / 2)
            else:
                break
        for i in range(1, 101):
            n = i
            if kronecker(n, Prime) == -1:
                break
        z = power_mod(ZZ(n),ZZ(q),ZZ(Prime))
        y = ZZ(z)
        r = ZZ(e)
        x = power_mod(ZZ(a),ZZ( (q-1)/2),ZZ( Prime))
        b = (a * x ** 2) % Prime
        x = (a * x) % Prime
        #for i in range(1, e + 1):
        for i in itertools.count(1):
            if b == 1:
                break
            else:
                for m in itertools.count(1):
                    t = power_mod(ZZ(b), ZZ(2^m) ,  ZZ(Prime))
                    if t == 1:
                        mm = m
                        break
                t = power_mod(ZZ(y), ZZ(2^(r - mm - 1)),ZZ(Prime))
                y = power_mod(ZZ(t), ZZ(2),ZZ(Prime))
                r = mm
                x = (x * t) % Prime
                b = (b * y) % Prime
        return(x)

︡5d13deb4-82f6-49a1-869e-544e7c43ff39︡{"done":true}
︠236c4656-e579-433e-8e97-d3abdb067738i︠
%html
<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between "lowerbound" and "upperbound". The lowerbound and the upperbound are positive random numbers.
︡eefda683-06b5-4a0c-89e8-4f730397080d︡{"done":true,"html":"<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between \"lowerbound\" and \"upperbound\". The lowerbound and the upperbound are positive random numbers."}
︠e6c93c60-6ba8-4bf5-b5db-7557b6152d72s︠
import itertools
def ECSearch (lowerbound, upperbound, Group):
    p = Group[2]
    a = Group[0]
    b = Group[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print("This is not an elliptic curve. ")
    else:
        for j in itertools.count(lowerbound):
            if j==lowerbound-1 or j>upperbound or j>upperbound:
                return "not found"
            j=j%p
            #print "test "+str(j)
            #print (kronecker(j ** 3 + a * j + b, p) )
            if kronecker(j ** 3 + a * j + b, p) == 1:
                x = (j ** 3 + a * j + b) % p
                var('z')
                #print ("Modsqrt({0}, {1})".format(x,p))
                #L = solve_mod(z ** 2 == x, p)
                L=TonSh(x,p)
                y = L
                #y = L[0][0]
                return([j,y])
︡4a532090-4001-45d1-ad59-99bf19bdfdc2︡{"done":true}
︠d52d71b7-d474-416a-b706-2e45d80067cai︠
%html
<strong>EXAMPLE:</strong> Find a point on the elliptic curve $y^2=x^3+2x+14\;mod\; 1291$ such that the first coordinate of the point is between (234,789).
︡f19f7aa3-bdde-4563-ba78-2a70915498c4︡{"done":true,"html":"<strong>EXAMPLE:</strong> Find a point on the elliptic curve $y^2=x^3+2x+14\\;mod\\; 1291$ such that the first coordinate of the point is between (234,789)."}
︠c14b49fb-0c25-4408-86ec-672b2ca4b03fs︠

P=ECSearch(234,789,[2,14,1291])
P
Q=ECSearch(345,891,[2,14,1291])
Q
︡4db3a99b-fc9f-4627-9205-d13e2748b8a4︡{"stdout":"[236, 1131]\n"}︡{"stdout":"[346, 749]\n"}︡{"done":true}
︠30b171c2-58ae-432a-b75a-5b8aed7511e1s︠
import sympy.ntheory
def FermatGreat (N):
    if sympy.ntheory.isprime(N):
        if N % 4 == 1:
            Z = two_squares(N)
            #print(Z)
            return(Z)
        else:
            print("{0} is prime, but {1} mod 4 = 3").format(N, N)
    else:
        print("{0} is not prime\n").format(N)

︡56e5b0d2-4e48-4f2e-8e1a-b7fd876e31de︡{"done":true}
︠0fac2b38-4c72-454a-8c4e-f5fcfd3cd06bs︠
FermatGreat(17)
︡bef23d8e-b055-4d54-9be8-f427dd40f6f5︡{"stdout":"(1, 4)\n"}︡{"done":true}
︠3c29af8f-3de6-4abb-83c6-e9538b32c55bi︠

%html
<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation.
︡2358f900-5d21-4047-bac4-3c125f016827︡{"done":true,"html":"<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation."}
︠bacf3b71-1c6f-4946-83fd-acfbfda8b94fs︠

︡40ecaf29-e252-4d5e-8cf6-74a859e3a2d3︡{"done":true}
︠d6ba2d47-db9c-4531-b62c-32bdefe2aa8ds︠
def ECAdd (Point1, Point2, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if(Point1!=[]):
        x1 = Point1[0]
        y1 = Point1[1]
    if(Point2!=[]):
        x2 = Point2[0]
        y2 = Point2[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print("This is not an elliptic curve. ")
    elif Point1 != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
        print("Point 1 is not on the elliptic curve. \n")
    elif Point2 != [] and y2 ** 2 % p != (x2 ** 3 + a * x2 + b) % p:
        print("Point 2 is not on the elliptic curve. \n")
    else:
        if Point1 == []:
            R = Point2
        elif Point2 == []:
            R = Point1
        else:
            if x1 == x2 and 0 == (y1 + y2) % p:
                R = []
            elif x1 == x2 and y1 == y2:
                R = ECDouble(Point1, Group)
            else:
                s = ((y1 - y2) / (x1 - x2)) % p
                x = (s ** 2 - x1 - x2) % p
                y = (s * (x1 - x) - y1) % p
                R = [x,y]
    return(R)
︡4c537c9e-7cf0-4f33-b616-ed60d01d7694︡{"done":true}
︠94607842-b684-4ada-b236-b216ffed1bf5i︠
%html
<strong>EXAMPLE:</strong> Find the sum of the points (236, 1131) and (346, 749) on the elliptic curve $y^2=x^3+2x+14\;mod\; 1291$.
︡da8347a9-db33-40fd-a762-874aa85b1883︡{"done":true,"html":"<strong>EXAMPLE:</strong> Find the sum of the points (236, 1131) and (346, 749) on the elliptic curve $y^2=x^3+2x+14\\;mod\\; 1291$."}
︠3f92ee06-5be1-43b6-b98e-198c42e207d5s︠

ECAdd([236,1131],[346,749],[2,14,1291])
︡4aec8ced-c87c-48e9-9eb1-92b1ce363a02︡{"stdout":"[1109, 774]\n"}︡{"done":true}
︠795080ab-d669-44e3-bd82-7d5e0cdd17aes︠

︡1e73427c-1b60-4639-b71c-6e3c11a7fc4e︡{"done":true}
︠2c75fed5-4477-4f1b-85e7-a17925d96a6cs︠
def ECDouble (Point, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if Point != []:
        x1 = Point[0]
        y1 = Point[1]
        if (4 * a ** 3 + 27 * b ** 2) % p == 0:
            print("This is not an elliptic curve. ")
        elif Point != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
            print("The point to double is not on the elliptic curve. ")
        elif y1 == 0:
            R = []
        else:
            s = mod((3 * x1 ** 2 + a) / y1 / 2, p)
            x = (s ** 2 - 2 * x1) % p
            y = (s * (x1 - x) - y1) % p
            R = [x,y]
    else:
        R = []
    return(R)
︡a2f48a6e-6672-4e46-a499-21e0a22cf948︡{"done":true}
︠908999b7-3584-4f01-8203-e3c5fe62cc1di︠
%html
<strong>EXAMPLE:</strong> Find the sum of the point (6,9) to itself i.e. $2\cdot(6,9)$ on the curve $y^2=x^3+x+2\;mod\; 13$.
︡7521532b-dc1b-4030-afaf-e57a41deb76e︡{"done":true,"html":"<strong>EXAMPLE:</strong> Find the sum of the point (6,9) to itself i.e. $2\\cdot(6,9)$ on the curve $y^2=x^3+x+2\\;mod\\; 13$."}
︠6494f087-8504-40fa-8a7f-86a14d892fefs︠

ECDouble([6,9],[1,2,13])
︡b5dfabde-2ad6-453f-b76b-9c094971030a︡{"stdout":"[2, 8]\n"}︡{"done":true}
︠9cd6481d-6bb8-4fbe-802d-5594417b0978i︠
%html
<strong>ECInverse</strong> is a procedure that finds the inverse of a point on a given elliptic curve using the elliptic curve operation.
︡39285021-1be2-4ce6-9474-e0ad2580ddcf︡{"done":true,"html":"<strong>ECInverse</strong> is a procedure that finds the inverse of a point on a given elliptic curve using the elliptic curve operation."}
︠aa74528c-8834-40a9-ae2d-c3a1b0e42632s︠
def ECInverse (Point, Group):
    if Point == []:
        return(Point)
    else:
        p = Group[2]
        x = Point[0]
        y = Point[1]
        return([x,(p - y) % p])
︡bc95a588-25bc-4a2d-8031-ecab7f1a9257︡{"done":true}
︠94078dcf-a4d7-46e1-aae4-556fece61f31i︠
%html
<strong>EXAMPLE:  </strong>
Find an inverse of the point (6,9) on the curve $y^2=x^3+x+2\;mod\; 13$.
︡4d27a388-8433-4bba-be9c-eebeeeb1bba3︡{"done":true,"html":"<strong>EXAMPLE:  </strong>\nFind an inverse of the point (6,9) on the curve $y^2=x^3+x+2\\;mod\\; 13$."}
︠1473006c-15e0-4465-8146-f8e2b3fb66a1s︠

ECInverse([6,9],[1,2,13])
︡b6dd3c3a-4923-4d6d-9947-75c898ae6eb6︡{"stdout":"[6, 4]\n"}︡{"done":true}
︠248b5919-f1b5-4029-ace3-e1a2a98d2caai︠
%html
<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation.
︡18b4f37e-91cf-4637-ba84-cac38e1450c1︡{"done":true,"html":"<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation."}
︠db90e0c4-4f08-428b-933d-8daa383a99d9s︠
def ECTimes (Point, scalar, Group):
    ECIdentity=[]
    if Point == ECIdentity or scalar == 0:
        return(ECIdentity)
    else:
        E=EllipticCurve(GF(Group[2]),[Group[0],Group[1]])
        EPoint = E(Point[0],Point[1])
        #print type(EPoint)
        cgret = scalar*EPoint
        #print cgret[0]
        if(cgret[0]==0 and cgret[1]==1 and cgret[2]==0):
            return ECIdentity
        return([cgret[0],cgret[1]])

︡bd523081-5ed4-41f4-8491-545e818e589b︡{"done":true}
︠f9f6a715-09b8-4db7-81ec-7905877e9601i︠
%html
<strong>EXAMPLE:</strong> Multiply the point (1,11) by 71 (add the point (1,11) to itself 71 times) on the curve $y^2=x^3+x+2\;mod\; 13$.
︡5ace523f-bdd1-4ee1-a10b-e0c0cc70c28b︡{"done":true,"html":"<strong>EXAMPLE:</strong> Multiply the point (1,11) by 71 (add the point (1,11) to itself 71 times) on the curve $y^2=x^3+x+2\\;mod\\; 13$."}
︠c5373666-181f-46a4-a5b4-0ec1f6498102s︠
ECTimes([1,11],71,[1,2,13])
G=[3,2,next_prime(11696876878768768767867687687687687678768768766876879878111)]
G ####### Elliptic Curve Group
g=ECSearch(2378, 892749287373,G)
g ####### base element
x=34239473274983742
x ####### private key
b=ECTimes(g,x,G)
︡328163d4-57fe-4255-bc66-7fc563786abe︡{"stdout":"[1, 2]\n"}︡{"stdout":"[3, 2, 11696876878768768767867687687687687678768768766876879878131]\n"}︡{"stdout":"[2378, 1882214856783904319500075249934504476880494552940632700484]\n"}︡{"stdout":"34239473274983742\n"}︡{"done":true}
︠fcc65658-00c8-4270-a732-81804c9fce6as︠

︡97621e3a-8d82-40a9-a22c-6b8485434cf6︡{"done":true}
︠54506cb7-0c5b-4b49-afdd-0a8bc3b1da6fs︠
def listproduct (primelist):
    x = primelist
    k = len(primelist)
    print(k)
    result = 1
    for i in range(0, k):
        result = result * x[i][0] ^ x[i][1] #op(1, op(i, x)) ** op(2, op(i, x))
    return(result)

def listptorder (pt, Group, factoredGpOrder):
    ECIdentity = []
    x = factoredGpOrder
    k = len(factoredGpOrder)
    orderlist = set([])
    result = listproduct(factoredGpOrder)
    for i in range(0, k ):
        p =  x[i][0] #op(1, op(i, x))
        a =  x[i][1] #op(2, op(i, x))
        ord = 0
        for j in range(0, a ):
            result = result / p
            test = ECTimes(pt, result, Group)
            if test != ECIdentity:
                result = result * p
                ord = ord + 1
        if 0 < ord:
            orderlist = orderlist | set([[p,ord]])
    return(convert(orderlist, list))


def primeHPSonEC (y,g,Group,p):
    newg = []#ECIdentity
    for j in range(0, p ):
        if y == newg:
            break
        newg = ECAdd(newg, g, Group)
    return(j)


def powerHPSonEC (y,g,og,p,a,Group):
    gp = ECTimes(g, og / p, Group)
    newog = og
    newy = y
    c = 0
    partx = 0
    for i in range(0, a ):
        newy = ECAdd(y, ECInverse(ECTimes(g, partx, Group), Group), Group)
        newog = newog / p
        ypower = ECTimes(newy, newog, Group)
        c = primeHPSonEC(ypower, gp, Group, p)
        partx = partx + c * p ** i
    return(partx)

def HPSonEC (y,g,Group,factoredGpOrder):
    X = 0
    Ord = 1
    K = listptorder(g, Group, factoredGpOrder)
    k = len(K)
    ordg = listproduct(K)
    for j in range(1, k + 1):
        p = K[j][0]#op(1, op(j, K))
        powerp = K[j][1]#op(2, op(j, K))
        partx = powerHPSonEC(y, g, ordg, p, powerp, Group)
        if j == 1:
            X = partx
            Ord = p ** powerp
        else:
            X = crt([X,partx], [Ord,p ** powerp])
            Ord = Ord * p ** powerp
    return(X)
︡9fb6e089-88c1-4a51-8156-0da9fd9471d2︡{"done":true}
︠12961394-fae6-4142-b4d8-96147d9f17cas︠

def ECEmbed (Message, gp, tolerance):
    p = ZZ(math.floor(gp[2] / (tolerance + 1)))
    M = ASCIIPad(Message, p)
    packets = len(M)
    #print packets
    pointlist = [0]*packets
    for j in range(0, packets ):
        N = M[j]
#        N:=(op(0,op(1,M))[j]);
        pointlist[j] = ECSearch(tolerance * N, tolerance * (N + 1) - 1, gp)
        #print pointlist
    return(pointlist)

def ECUnembed (pointlist, tolerance):
    #print pointlist
    k = len(pointlist)
    paddedasciilist=[0]*k
    for j in range(0, k ):
        #print type(pointlist[j][0]/tolerance)
        #print pointlist
        #print paddedasciilist
        #print "test this "+str((pointlist[j][0]/tolerance))
        #.floor() returns correct while math.floor() does not work
        #print (pointlist[j][0])
        pointlist[j][0]=ZZ(pointlist[j][0])
        toType = ZZ(QQ((pointlist[j][0])/tolerance).floor())
#        typed = type(toType)
#        print typed
        paddedasciilist[j] = ((pointlist[j][0])/tolerance).floor()
    returnStr = ""
    for paddedItem in paddedasciilist:
        #print paddedItem
        buffer = ASCIIDepad(paddedItem)
        returnStr+= buffer
        #print buffer
    return returnStr


import math

def ASCIIPad(Mess,Mod):
    K = []
    for letter in Mess:
        K.append(ord(letter))
    L = Mod.ndigits()
    l = len(K)
    y = ZZ(math.floor(L/3))
    count = 0
    padded = []
    buffer = ""
    for numChar in K:
        numChar+=100
        buffer+=str(numChar)
        count+=1
        if count==y:
            padded.append(ZZ(buffer))
            buffer = ""
            count = 0
    if len(buffer)>0:
        padded.append(ZZ(buffer))
    return padded


def ASCIIDepad(Number):
    N = "";
    #Number = ZZ(Number[0])
    n = Number.ndigits() % 3;
    if (n > 0):
        print("This is not a padded ASCII string\n");
    else:
        L = [((Number - (Number % (1000^i)))/1000^i)%1000 - 100 for i in range(Number.ndigits()/3)];
        for i in range(Number.ndigits()/3):
            #print L[i]
            N = chr(L[i]) + N;

    return(N);

︡196204ec-710c-4f04-aef2-f4f36636173f︡{"done":true}
︠cc7f40ba-7fdd-4db7-a081-ac91428bb53fi︠
%html
<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between "lowerbound" and "upperbound". The lowerbound and the upperbound are positive random numbers.
︡dedfcc80-9fb7-4119-9ef7-122d3a99801f︡︡{"done":true,"html":"<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between \"lowerbound\" and \"upperbound\". The lowerbound and the upperbound are positive random numbers."}
︠aa2f2834-10ab-491d-a5b9-5a2bbdfa3884s︠
import itertools
def ECSearch (lowerbound, upperbound, Group):
    p = Group[2]
    a = Group[0]
    b = Group[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print("This is not an elliptic curve. ")
    else:
        for j in itertools.count(lowerbound):
            if j==lowerbound-1 or j>upperbound or j>upperbound:
                return("not found")
            j=j%p
            #print "test "+str(j)
            #print (kronecker(j ** 3 + a * j + b, p) )
            if kronecker(j ** 3 + a * j + b, p) == 1:
                x = (j ** 3 + a * j + b) % p
                var('z')
                #print("stuck here")
                #print ("Modsqrt({0}, {1})".format(x,p))
                #L = solve_mod(z ** 2 == x, p)
                L=TonSh(x,p)
                y = L
                #y = L[0][0]
                return([j,y])
︡e14eae22-388e-4ad4-b0ab-e63c3f5aa996︡{"done":true}
︠ab1610d6-707e-41a5-9d5c-f6c7cefe6c6ds︠
#pr=next_prime(17416141711513452221)
#Ec=[21,3,pr]
#pt=[1,5]
#pk=834756
#rn = 324765348769
#G1 = EllipticCurve(GF(pr,'a'),[21,3])
#p1 = G1(1,5)
#pk = pk*p1
#rp = rn*p1
#Go = G1.order()
##fgo = factor(Go)
##print(pr, pk, Go, fgo)
#Message = "Deposit $50 in account number 23456789"
#EMessage = ECEmbed(Message,G1,50)
#EMessage
##AM = ASCIIPad(Message,Go)
##AM
##D=ECSearch(29,31,E)
##Q=ECSearch(34,36,E)
##D
##Q
##
︡eae56a9b-f38d-45ed-a76e-08e3909fd2ec︡{"done":true}
︠e4ed0228-2298-4ca4-b6f0-e5baffb93264︠

︡559bb2b7-f747-4d7f-b4fd-ab5618339851︡{"stdout":"2398456394587698346003\n"}︡{"done":true}
︠35a73da4-d03b-430a-8a5a-93b203947b3b︠










